// amqmskyn.cpp : Defines the entry point for the console application.
// |                  Licensed Materials - Property of IBM              |
// |                  SupportPac MC73                                   |
// |                                                                    |
// |                  (c) Copyright IBM Corp. 1999                      |
// |                                                                    |
// |                  All rights reserved.                              |
// |                                                                    |
// |                  U.S. Government Users Restricted Rights - use,    |
// |                  duplication or disclosure restricted by GSA       |
// |                  ADP Schedule Contract with IBM Corp.              |
//

#include <stdio.h>
#include <windows.h>

#include <aclapi.h>
#include <lm.h>
#include <lmerr.h>
#include <mbstring.h>

#define irgMQ_KEY                       "Software\\IBM\\MQSeries"
#define irgP000_CONFIGURATION_ROOT      "\\CurrentVersion\\Configuration"

#define PSZ_USER_MQM          "mqm"
#define PSZ_DOMAIN_USER_MQM   "Domain mqm"
#define MACHINE_KEY           "MACHINE"

#define MAXSTRING	1024

extern BOOL isuSetOwner(CHAR *pszObjectName, SE_OBJECT_TYPE ObjectType);
extern BOOL	isuChangeKeySecurity(CHAR        *pszKey,
                          BOOL        fDiscardOld,
                          ACCESS_MODE WorldAccessMode,
                          DWORD       WorldAccessPerms,
                          DWORD       WorldInherit);
extern ULONG isuIsDomainController(VOID);

static SID_IDENTIFIER_AUTHORITY SidAuthNT    = SECURITY_NT_AUTHORITY;
static SID_IDENTIFIER_AUTHORITY SidAuthWorld = SECURITY_WORLD_SID_AUTHORITY;

static   PSID CreateWorldSid(PSID pSid);
static   PSID CreateAdminSid(PSID pSid);
static   PSID CreateMqmSid(PSID pSid);
static   PSID CreateSystemSid(PSID pSid);
static   PSID CreateUserSid(PSID pSid);

static PSID             pSIDAdmin = NULL;
static PSID             pSIDSystem = NULL;
static PSID             pSIDWorld = NULL;
static PSID             pSIDMqm = NULL;

extern void SetMQMKeySecurity(char*  lpszKey);
extern void isuWriteLog(char* lpszText);

int main(int argc, char* argv[])
{
	char szKey[MAXSTRING];
    sprintf(szKey, "%s%s", irgMQ_KEY, irgP000_CONFIGURATION_ROOT);

	// Process this key, and then recurse through all subkeys
	SetMQMKeySecurity(szKey);

	return 0;
}


void isuWriteLog(char* lpszText)
{
	printf("%s\n",lpszText);
}

/**********************************************************************/
/*                                                                    */
/* Recursive function to set security on a key and all it's subkeys   */
/*                                                                    */
/**********************************************************************/
void SetMQMKeySecurity(char* lpszKey)
{
	CHAR csSubKey[MAXSTRING];

 	isuSetOwner(lpszKey, SE_REGISTRY_KEY);
	isuChangeKeySecurity(lpszKey,
		FALSE,
		NOT_USED_ACCESS,
		0, 0);

	HKEY hKeyRoot;
	if(RegOpenKey(HKEY_LOCAL_MACHINE, lpszKey, &hKeyRoot) == ERROR_SUCCESS)
	{

		short nIndex = 0;
		// recurse for all subkeys under this key
		do
		{
			char csName[MAXSTRING];	
			DWORD dwSizeName = sizeof(csName);
			memset(csName, 0, dwSizeName);

			long lRes = RegEnumKey(hKeyRoot, (DWORD)nIndex, csName, dwSizeName);
			if(lRes == ERROR_SUCCESS)
			{
				dwSizeName = sizeof(csName);
				strcpy(csSubKey, lpszKey);
				strcat(csSubKey, "\\");
				strcat(csSubKey, csName);

				SetMQMKeySecurity(csSubKey);

				nIndex++;
			}
			else
				nIndex = -1;	// enumerate failed - clean up and finish this key

		}
		while(nIndex > 0);

		RegCloseKey(hKeyRoot);
	}
}

/**********************************************************************/
/*                                                                    */
/**********************************************************************/
BOOL isuChangeKeySecurity(CHAR        *pszKey,
                          BOOL        fDiscardOld,
                          ACCESS_MODE WorldAccessMode,
                          DWORD       WorldAccessPerms,
                          DWORD       WorldInherit)
{
  /********************************************************************/
  /* Local Variables                                                  */
  /********************************************************************/
  BOOL                  fSuccess = TRUE;
  CHAR                  szText[MAXSTRING];
  CHAR                  szRegKey[_MAX_PATH];
  DWORD                 dwSecRc;
  PSECURITY_DESCRIPTOR  pDesc = NULL;
  PACL                  pAclOld = NULL;
  PACL                  pAclNew = NULL;
  EXPLICIT_ACCESS       ExplicitAccess[4];
  ULONG                 ulAceCount;

  /********************************************************************/
  /* Create a SYSTEM SID                                              */
  /********************************************************************/
  pSIDSystem = CreateSystemSid(pSIDSystem);

  /********************************************************************/
  /* Create a world SID                                               */
  /********************************************************************/
  pSIDWorld = CreateWorldSid(pSIDWorld);

  /********************************************************************/
  /* Create a SID for Administrators group                            */
  /********************************************************************/
  pSIDAdmin = CreateAdminSid(pSIDAdmin);

  /********************************************************************/
  /* Create a SID for the local mqm group                             */
  /********************************************************************/
  pSIDMqm = CreateMqmSid(pSIDMqm);

  /********************************************************************/
  /* If SID creation has failed, quit                                 */
  /********************************************************************/
  if (!pSIDWorld || !pSIDAdmin || !pSIDMqm || !pSIDSystem)
  {
    /******************************************************************/
    /*                                                                */
    /******************************************************************/
    fSuccess = FALSE;
    sprintf(szText, "***Create SID error, file: %s", szRegKey);
    isuWriteLog(szText);
  }

  /********************************************************************/
  /*                                                                  */
  /********************************************************************/
  if (fSuccess)
  {
    /******************************************************************/
    /* Format the key with the "HKEY_LOCAL_MACHINE" literal.          */
    /******************************************************************/
    sprintf(szRegKey, "MACHINE\\%s", pszKey);

    /******************************************************************/
    /* Get the existing ACL (if any).                                 */
    /******************************************************************/
    if (!fDiscardOld)
    {
      dwSecRc = GetNamedSecurityInfo(szRegKey,
                                     SE_REGISTRY_KEY,
                                     DACL_SECURITY_INFORMATION,
                                     NULL,
                                     NULL,
                                     &pAclOld,
                                     NULL,
                                     &pDesc);
      if (ERROR_SUCCESS != dwSecRc)
      {
        pAclOld = NULL;                  // We don't have an ACL
        if (ERROR_FILE_NOT_FOUND != dwSecRc
            && ERROR_PATH_NOT_FOUND != dwSecRc)
        {
          fSuccess = FALSE;
          sprintf(szText, "***Error %d getting security info '%s'",
                          dwSecRc, szRegKey);
          isuWriteLog(szText);
        }
      }
    }

    /******************************************************************/
    /*                                                                */
    /******************************************************************/
    if (fSuccess)
    {
      /****************************************************************/
      /* Initialise an EXPLICIT_ACCESS structure for administrators.  */
      /****************************************************************/
      ZeroMemory(&ExplicitAccess[0], sizeof(EXPLICIT_ACCESS));
      BuildExplicitAccessWithName(&ExplicitAccess[0],
                                  "Admin", // ptstrName replaced shortly
                                   GENERIC_ALL,
                                   SET_ACCESS,
                                   CONTAINER_INHERIT_ACE
                                   | OBJECT_INHERIT_ACE);
      ExplicitAccess[0].Trustee.TrusteeForm = TRUSTEE_IS_SID;
      ExplicitAccess[0].Trustee.ptstrName = (LPTSTR)pSIDAdmin;

      /****************************************************************/
      /* Initialise an EXPLICIT_ACCESS structure for Mqm.             */
      /****************************************************************/
      ZeroMemory(&ExplicitAccess[1], sizeof(EXPLICIT_ACCESS));
      BuildExplicitAccessWithName(&ExplicitAccess[1],
                                  "MQM", // ptstrName replaced shortly
                                   GENERIC_ALL,
                                   SET_ACCESS,
                                   CONTAINER_INHERIT_ACE
                                   | OBJECT_INHERIT_ACE);
      ExplicitAccess[1].Trustee.TrusteeForm = TRUSTEE_IS_SID;
      ExplicitAccess[1].Trustee.ptstrName = (LPTSTR)pSIDMqm;

      /****************************************************************/
      /* Initialise an EXPLICIT_ACCESS structure for system.          */
      /****************************************************************/
      ZeroMemory(&ExplicitAccess[2], sizeof(EXPLICIT_ACCESS));
      BuildExplicitAccessWithName(&ExplicitAccess[2],
                                  "System", // ptstrName replaced shortly
                                   GENERIC_ALL,
                                   SET_ACCESS,
                                   CONTAINER_INHERIT_ACE
                                   | OBJECT_INHERIT_ACE);
      ExplicitAccess[2].Trustee.TrusteeForm = TRUSTEE_IS_SID;
      ExplicitAccess[2].Trustee.ptstrName = (LPTSTR)pSIDSystem;

      /****************************************************************/
      /* Initialise an EXPLICIT_ACCESS structure for world.           */
      /****************************************************************/
      if (NOT_USED_ACCESS == WorldAccessMode)
      {
        ulAceCount = 3;
      }
      else
      {
        ulAceCount = 4;
        ZeroMemory(&ExplicitAccess[3], sizeof(EXPLICIT_ACCESS));
        BuildExplicitAccessWithName(&ExplicitAccess[3],
                                    "World", // ptstrName replaced shortly
                                     WorldAccessPerms,
                                     WorldAccessMode,
                                     CONTAINER_INHERIT_ACE
                                     | OBJECT_INHERIT_ACE);
        ExplicitAccess[3].Trustee.TrusteeForm = TRUSTEE_IS_SID;
        ExplicitAccess[3].Trustee.ptstrName = (LPTSTR)pSIDWorld;
      }

      /****************************************************************/
      /* Merge a new ACE into the old ACL, creating a new ACL.        */
      /****************************************************************/
      dwSecRc = SetEntriesInAcl(ulAceCount,
                                &ExplicitAccess[0],
                                pAclOld, &pAclNew);
      if (ERROR_SUCCESS != dwSecRc)
      {
        fSuccess = FALSE;
        sprintf(szText, "***Error %d merging ACLs on '%s'",
                        dwSecRc, szRegKey);
        isuWriteLog(szText);
      }

      /****************************************************************/
      /* Set the updated security info                                */
      /****************************************************************/
      if (fSuccess)
      {
        dwSecRc = SetNamedSecurityInfo(szRegKey,
                                       SE_REGISTRY_KEY,
                                       DACL_SECURITY_INFORMATION,
                                       NULL,
                                       NULL,
                                       pAclNew,
                                       NULL);
        if (dwSecRc == ERROR_ACCESS_DENIED)
        {
          /************************************************************/
          /* If we have a security problem even though we're an       */
          /* administrator, it's probably because we're not the       */
          /* owner and that administrators have no access.            */
          /* So grab ownership and try again...                       */
          /************************************************************/
          isuSetOwner(pszKey, SE_REGISTRY_KEY);
          dwSecRc = SetNamedSecurityInfo(szRegKey,
                                         SE_REGISTRY_KEY,
                                         DACL_SECURITY_INFORMATION,
                                         NULL,
                                         NULL,
                                         pAclNew,
                                         NULL);
        }
        if (ERROR_SUCCESS != dwSecRc)
        {
          fSuccess = FALSE;
          sprintf(szText, "***Error %d setting key security on '%s'",
                          dwSecRc, szRegKey);
          isuWriteLog(szText);
        }
        else
        {
          sprintf(szText, "Set key security '%s'", pszKey);
          isuWriteLog(szText);
        }
      }
    }

    /******************************************************************/
    /* Free control blocks we allocated.                              */
    /******************************************************************/
    if (pAclNew)
    {
      LocalFree((HLOCAL)pAclNew);
    }

    if (pDesc)
    {
      LocalFree((HLOCAL)pDesc);
    }
  }

  /********************************************************************/
  /* Return to caller                                                 */
  /********************************************************************/
  return(fSuccess);
}

/**********************************************************************/
/* Function       : isuSetOwner()                                     */
/**********************************************************************/
/* Description    :                                                   */
/* Params In      :                                                   */
/* Params Out     : None                                              */
/* Params In/Out  : None                                              */
/* Returns        :                                                   */
/* Side effects   :                                                   */
/* Notes          :                                                   */
/*                                                                    */
/**********************************************************************/
BOOL   isuSetOwner(CHAR *pszObjectName, SE_OBJECT_TYPE ObjectType)
{
  /********************************************************************/
  /* Local variables                                                  */
  /********************************************************************/
  BOOL                  fSuccess;
  DWORD                 dwSecRc;
  CHAR                  szText[MAXSTRING];
  BOOL                  fPrivOk;
  HANDLE                ProcessToken = NULL;
  TOKEN_PRIVILEGES      NewPriv;
  LUID                  BackupLuid;
  static BOOL           fGotOwnerPriv = FALSE;
  CHAR                  szLocalObjectName[_MAX_PATH];

  /********************************************************************/
  /* Initialisation                                                   */
  /********************************************************************/
  fSuccess = TRUE;

  /********************************************************************/
  /*                                                                  */
  /********************************************************************/
  _mbscpy((unsigned char*)szLocalObjectName, (unsigned char*)pszObjectName);
  if (SE_REGISTRY_KEY == ObjectType)
  {
    if (0 != _mbsnicmp((unsigned char*)pszObjectName, (unsigned char*)MACHINE_KEY, _mbslen((unsigned char*)MACHINE_KEY)))
    {
      sprintf(szLocalObjectName, "%s\\%s", MACHINE_KEY, pszObjectName);
    }
  }

  /********************************************************************/
  /* Create a SID for Administrators group (if not already done).     */
  /********************************************************************/
  pSIDAdmin = CreateAdminSid(pSIDAdmin);

  /********************************************************************/
  /* If we've got a SID, get ownership of the object.                 */
  /********************************************************************/
  if (pSIDAdmin)
  {
    /******************************************************************/
    /* We need SE_TAKE_OWNERSHIP_NAME privilege to take ownership     */
    /* of secure objects, so let's get it...                          */
    /******************************************************************/
    if (!fGotOwnerPriv)
    {
      fPrivOk = TRUE;
      if (!LookupPrivilegeValue(NULL, SE_TAKE_OWNERSHIP_NAME, &BackupLuid))
      {
        sprintf(szText, "LookupPrivilegeValue error %d", GetLastError());
        isuWriteLog(szText);
        fPrivOk = FALSE;
      }

      if (fPrivOk)
      {
        if (!OpenProcessToken(GetCurrentProcess(),
                              TOKEN_ADJUST_PRIVILEGES
                              | TOKEN_QUERY,
                              &ProcessToken))
        {
          sprintf(szText, "OpenProcessToken error %d", GetLastError());
          isuWriteLog(szText);
          fPrivOk = FALSE;
        }
      }

      if (fPrivOk)
      {
        NewPriv.PrivilegeCount = 1;
        NewPriv.Privileges[0].Luid = BackupLuid;
        NewPriv.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
        if (!AdjustTokenPrivileges(ProcessToken,
                                   FALSE,
                                   &NewPriv,
                                   0,
                                   NULL,
                                   NULL))
        {
          sprintf(szText, "AdjustTokenPrivileges error %d", GetLastError());
          isuWriteLog(szText);
          fPrivOk = FALSE;
        }
      }

      if (fPrivOk)
      {
        fGotOwnerPriv = TRUE;
        isuWriteLog("SE_TAKE_OWNERSHIP_NAME privilege acquired");
      }
    }

    /******************************************************************/
    /* Now grab ownership. Ignore errors; something will fail later.  */
    /******************************************************************/
    dwSecRc = SetNamedSecurityInfo(szLocalObjectName,
                                   ObjectType,
                                   OWNER_SECURITY_INFORMATION,
                                   pSIDAdmin,
                                   NULL,
                                   NULL,
                                   NULL);
    if (ERROR_SUCCESS == dwSecRc)
    {
      sprintf(szText, "Set ownership on '%s'", szLocalObjectName);
      isuWriteLog(szText);
    }
    else
    {
      if (ERROR_FILE_NOT_FOUND != dwSecRc
          && ERROR_PATH_NOT_FOUND != dwSecRc)
      {
        sprintf(szText, "Error %d setting ownership on '%s'",
                        dwSecRc, szLocalObjectName);
        isuWriteLog(szText);
      }
    }
  }

  /********************************************************************/
  /* Return to caller                                                 */
  /********************************************************************/
  return(fSuccess);
}

/**********************************************************************/
/* Create a SID for the System group.                                 */
/**********************************************************************/
PSID  CreateSystemSid(PSID pSidIn)
{
  /********************************************************************/
  /* Local variables                                                  */
  /********************************************************************/
  LONG                     error;
  CHAR                     szText[MAXSTRING];
  PSID                     pSid;

  /********************************************************************/
  /* Create a SID for system group (if not already done).             */
  /********************************************************************/
  if (pSidIn)
  {
    pSid = pSidIn;                     // Already done
  }
  else
  {
    if (!AllocateAndInitializeSid(&SidAuthNT,
                                  1,
                                  SECURITY_LOCAL_SYSTEM_RID,
                                  0, 0, 0, 0, 0, 0, 0,
                                  &pSid))
    {
      pSid = NULL;
      error = GetLastError();
      sprintf(szText, "Error %d creating system SID", error);
      isuWriteLog(szText);
    }
  }

  /********************************************************************/
  /* Return to caller                                                 */
  /********************************************************************/
  return(pSid);
}

/**********************************************************************/
/* Create a world SID                                                 */
/**********************************************************************/
PSID  CreateWorldSid(PSID pSidIn)
{
  /********************************************************************/
  /* Local variables                                                  */
  /********************************************************************/
  LONG                     error;
  CHAR                     szText[MAXSTRING];
  PSID                     pSid;

  /********************************************************************/
  /* Create a SID for everybody (if not already done).                */
  /********************************************************************/
  if (pSidIn)
  {
    pSid = pSidIn;                     // Already done
  }
  else
  {
    if (!AllocateAndInitializeSid(&SidAuthWorld,
                                  1,
                                  SECURITY_WORLD_RID,
                                  0, 0, 0, 0, 0, 0, 0,
                                  &pSid))
    {
      pSid = NULL;
      error = GetLastError();
      sprintf(szText, "Error %d creating world SID", error);
      isuWriteLog(szText);
    }
  }

  /********************************************************************/
  /* Return to caller                                                 */
  /********************************************************************/
  return(pSid);
}

/**********************************************************************/
/* Create a SID for the Administrators group.                         */
/**********************************************************************/
PSID  CreateAdminSid(PSID pSidIn)
{
  /********************************************************************/
  /* Local variables                                                  */
  /********************************************************************/
  LONG                     error;
  CHAR                     szText[MAXSTRING];
  PSID                     pSid;

  /********************************************************************/
  /* Create a SID for everybody (if not already done).                */
  /********************************************************************/
  if (pSidIn)
  {
    pSid = pSidIn;                     // Already done
  }
  else
  {
    if (!AllocateAndInitializeSid(&SidAuthNT,
                                  2,
                                  SECURITY_BUILTIN_DOMAIN_RID,
                                  DOMAIN_ALIAS_RID_ADMINS,
                                  0, 0, 0, 0, 0, 0,
                                  &pSid))
    {
      pSid = NULL;
      error = GetLastError();
      sprintf(szText, "Error %d creating world SID", error);
      isuWriteLog(szText);
    }
  }

  /********************************************************************/
  /* Return to caller                                                 */
  /********************************************************************/
  return(pSid);
}


/**********************************************************************/
/* Create a SID for the logged-on user.                               */
/**********************************************************************/
PSID  CreateUserSid(PSID pSidIn)
{
  /********************************************************************/
  /* Local variables                                                  */
  /********************************************************************/
  LONG         error;
  CHAR         szText[MAXSTRING];
  PSID         pSid;
  BOOL         fSuccess = TRUE;
  DWORD        SIDUserLen = 64;        // Initial length for SID
  CHAR         RefDomainName[24];
  DWORD        RefDomainNameLen=sizeof(RefDomainName);
  SID_NAME_USE SIDNameUse;
  CHAR         szUserName[MAXSTRING];
  DWORD        dwUserSize;

  /********************************************************************/
  /* Create a SID for the user (if not already done).                 */
  /********************************************************************/
  if (pSidIn)
  {
    pSid = pSidIn;                     // Already done
  }
  else
  {
    if (!(pSid = (PSID)malloc(SIDUserLen)))
    {
      fSuccess = FALSE;
      sprintf(szText, "***Error allocating User SID buffer (1)", error);
      isuWriteLog(szText);
    }

    if (fSuccess)
    {
      dwUserSize = sizeof(szUserName);
      fSuccess = GetUserName(szUserName, &dwUserSize);
      if (!fSuccess)
      {
        error = GetLastError();
        sprintf(szText, "Can't get user name %d", error);
        isuWriteLog(szText);
      }
    }

    if (fSuccess)
    {
      RefDomainNameLen=sizeof(RefDomainName);
      fSuccess = LookupAccountName(0,
                                   szUserName,
                                   pSid,
                                   &SIDUserLen,
                                   RefDomainName,
                                   &RefDomainNameLen,
                                   &SIDNameUse);
      if (!fSuccess)
      {
        error = GetLastError();

        /*************************************************************/
        /* If SIDBuffer too small, malloc one and try again          */
        /*************************************************************/
        if (ERROR_INSUFFICIENT_BUFFER == error)
        {
          if (!(pSid = (PSID)malloc(SIDUserLen)))
          {
            fSuccess = FALSE;
            sprintf(szText, "***Error allocating user SID buffer (2)", error);
            isuWriteLog(szText);
          }

          fSuccess = LookupAccountName(0,
                                       szUserName,
                                       pSid,
                                       &SIDUserLen,
                                       RefDomainName,
                                       &RefDomainNameLen,
                                       &SIDNameUse);
          if (!fSuccess)
          {
            error = GetLastError();
            sprintf(szText, "Error %d looking up user account name (2)", error);
            isuWriteLog(szText);
          }
        }
        else
        {
          sprintf(szText, "Error %d looking up user account (1)", error);
          isuWriteLog(szText);
        }
      }
    }

    /******************************************************************/
    /* Free the SID buffer if we failed.                              */
    /******************************************************************/
    if (!fSuccess)
    {
      if (pSid) free(pSid);
      pSid = NULL;
    }
  }

  /********************************************************************/
  /* Return to caller                                                 */
  /********************************************************************/
  return(pSid);
}


/**********************************************************************/
/* Create a SID for the mqm group.                                    */
/**********************************************************************/
PSID  CreateMqmSid(PSID pSidIn)
{
  /********************************************************************/
  /* Local variables                                                  */
  /********************************************************************/
  LONG   error;
  CHAR   szText[MAXSTRING];
  PSID   pSid;
  BOOL   fSuccess = TRUE;
  DWORD  SIDMqmLen = 64;               // Initial length for SID
  CHAR   RefDomainName[24];
  DWORD  RefDomainNameLen=sizeof(RefDomainName);
  SID_NAME_USE SIDNameUse;
  LPBYTE pDcName = NULL;
  LONG   lDcType = 0L;
  CHAR   szDcName[MAXSTRING];
  CHAR*  pszDcName = NULL;
  ULONG  ulNetRc = 0;

  /********************************************************************/
  /* Create a SID for mqm group (if not already done).                */
  /********************************************************************/
  if (pSidIn)
  {
    pSid = pSidIn;                     // Already done
  }
  else
  {
    if (!(pSid = (PSID)malloc(SIDMqmLen)))
    {
      fSuccess = FALSE;
      sprintf(szText, "***Error allocating MQM SID buffer (1)");
      isuWriteLog(szText);
    }

    if (fSuccess)
    {
	  lDcType = isuIsDomainController();

	  if (SV_TYPE_DOMAIN_BAKCTRL == lDcType)
	  {
		  ulNetRc = NetGetDCName(NULL, NULL, &pDcName);
		  if (NERR_Success == ulNetRc)
		  {
			WideCharToMultiByte(CP_ACP, 0, (LPCWSTR)pDcName, -1,
								szDcName, sizeof(szDcName), NULL, NULL);
			pszDcName = &szDcName[0];
		  }
		  else
		  {
			sprintf(szText, "***Error CreateMqmSid() NetGetDCName rc %d", ulNetRc);
	  	    isuWriteLog(szText);
		  }
	  }

      RefDomainNameLen=sizeof(RefDomainName);
      fSuccess = LookupAccountName(pszDcName,
                                   PSZ_USER_MQM,
                                   pSid,
                                   &SIDMqmLen,
                                   RefDomainName,
                                   &RefDomainNameLen,
                                   &SIDNameUse);
      if (!fSuccess)
      {
        error = GetLastError();

        /*************************************************************/
        /* If SIDBuffer too small, malloc one and try again          */
        /*************************************************************/
        if (ERROR_INSUFFICIENT_BUFFER == error)
        {
          if (!(pSid = (PSID)malloc(SIDMqmLen)))
          {
            fSuccess = FALSE;
            sprintf(szText, "***Error allocating MQM SID buffer (2)", error);
            isuWriteLog(szText);
          }

          fSuccess = LookupAccountName(pszDcName,
                                       PSZ_USER_MQM,
                                       pSid,
                                       &SIDMqmLen,
                                       RefDomainName,
                                       &RefDomainNameLen,
                                       &SIDNameUse);
          if (!fSuccess)
          {
            error = GetLastError();
            sprintf(szText, "***Error %d looking up MQM account name (2)", error);
            isuWriteLog(szText);
          }
        }
        else
        {
          sprintf(szText, "***Error %d looking up MQM account (1)", error);
          isuWriteLog(szText);
        }
      }
    }

    /******************************************************************/
    /* Free the SID buffer if we failed.                              */
    /******************************************************************/
    if (!fSuccess)
    {
      if (pSid) free(pSid);
      pSid = NULL;
    }
  }

  if (pDcName) NetApiBufferFree(pDcName);

  /********************************************************************/
  /* Return to caller                                                 */
  /********************************************************************/
  return(pSid);
}

/**********************************************************************/
/* Function       : isuIsDomainController()                           */
/**********************************************************************/
/* Description    : Determines if executing on a domain controller    */
/* Params In      : None                                              */
/* Params Out     : None                                              */
/* Params In/Out  : None                                              */
/* Returns        : [0,8,16] 0 = NO, 8 = PDC, 16 = BDC                */
/* Side effects   :                                                   */
/* Notes          :                                                   */
/*                                                                    */
/**********************************************************************/
ULONG isuIsDomainController(VOID)
{
  LONG               ulDcType = 0;
  ULONG              ulNetRc = 0;
  PSERVER_INFO_101   pServerInformation = NULL;
  CHAR               szText[MAXSTRING];
  LPBYTE             pDcName = NULL;
  CHAR               szDcName[MAXSTRING];

  /********************************************************************/
  /* Check to see if we are a PDC or BDC.                             */
  /********************************************************************/
  ulNetRc = NetServerGetInfo(NULL,
                             101,
                             (LPBYTE *)&pServerInformation);
  if (NERR_Success == ulNetRc)
  {
    if (pServerInformation->sv101_type & SV_TYPE_DOMAIN_CTRL)
    {
      ulDcType = SV_TYPE_DOMAIN_CTRL;
      isuWriteLog("Executing on a primary domain controller");
    }
    else
    if (pServerInformation->sv101_type & SV_TYPE_DOMAIN_BAKCTRL)
    {
      ulDcType = SV_TYPE_DOMAIN_BAKCTRL;
      isuWriteLog("Executing on a backup domain controller");
      ulNetRc = NetGetDCName(NULL, NULL, &pDcName);
      if (NERR_Success == ulNetRc)
      {
        WideCharToMultiByte(CP_ACP, 0, (LPCWSTR)pDcName, -1,
                            szDcName, sizeof(szDcName), NULL, NULL);
        sprintf(szText, "Primary controller is '%s'", szDcName);
      }
      else
      {
        sprintf(szText, "NetGetDCName rc %d", ulNetRc);
      }
      isuWriteLog(szText);
    }
    else
    {
      isuWriteLog("Not executing on a domain controller");
    }
  }
  else
  {
    sprintf(szText, "NetServerGetInfo rc %d", ulNetRc);
    isuWriteLog(szText);
  }

  /********************************************************************/
  /*                                                                  */
  /********************************************************************/
  if (pServerInformation) NetApiBufferFree(pServerInformation);
  if (pDcName) NetApiBufferFree(pDcName);

  /********************************************************************/
  /* Return to caller                                                 */
  /********************************************************************/
  return(ulDcType);
}
